package com.fzm.admin.listener;

import cn.fzm.chain.simplesdk.client.ParaChainClient;
import cn.hutool.core.date.DateUtil;
import cn.hutool.core.lang.TypeReference;
import cn.hutool.http.HttpUtil;
import cn.hutool.json.JSONUtil;
import com.fzm.common.config.RabbitMQConfig;
import com.fzm.common.entity.CopyrightApply;
import com.fzm.common.entity.Nft;
import com.fzm.common.entity.dto.CopyrightQueryRequest;
import com.fzm.common.entity.dto.CopyrightQueryResponse;
import com.fzm.common.entity.dto.EvidenceHashMessage;
import com.fzm.common.entity.vo.NftVo;
import com.fzm.common.enums.CopyrightApplyState;
import com.fzm.common.properties.CopyrightProperties;
import com.fzm.common.service.CopyrightApplyService;
import com.fzm.common.service.NftService;
import com.fzm.common.utils.CopyrightSignUtil;
import com.fzm.common.utils.JsonUtil;
import com.rabbitmq.client.Channel;
import lombok.extern.slf4j.Slf4j;
import org.apache.commons.lang3.StringUtils;
import org.springframework.amqp.core.Message;
import org.springframework.amqp.rabbit.annotation.RabbitListener;
import org.springframework.amqp.rabbit.core.RabbitTemplate;
import org.springframework.beans.factory.annotation.Value;
import org.springframework.stereotype.Component;

import javax.annotation.Resource;
import java.io.IOException;
import java.util.Map;
import java.util.TreeMap;

/**
 * @author tangtuo
 * @date 2021/12/23 14:56
 */
@Slf4j
@Component
public class CopyrightLister {

    @Resource
    private CopyrightProperties copyrightProperties;

    @Resource
    private RabbitTemplate rabbitTemplate;

    @Resource
    private CopyrightApplyService copyrightApplyService;

    @Resource
    private ParaChainClient paraChainClient;

    @Resource
    private NftService nftService;

    private static String abi = "[{\"inputs\":[],\"stateMutability\":\"nonpayable\",\"type\":\"constructor\"},{\"anonymous\":false,\"inputs\":[{\"indexed\":true,\"internalType\":\"address\",\"name\":\"account\",\"type\":\"address\"},{\"indexed\":true,\"internalType\":\"address\",\"name\":\"operator\",\"type\":\"address\"},{\"indexed\":false,\"internalType\":\"bool\",\"name\":\"approved\",\"type\":\"bool\"}],\"name\":\"ApprovalForAll\",\"type\":\"event\"},{\"anonymous\":false,\"inputs\":[{\"indexed\":true,\"internalType\":\"address\",\"name\":\"operator\",\"type\":\"address\"},{\"indexed\":true,\"internalType\":\"address\",\"name\":\"from\",\"type\":\"address\"},{\"indexed\":true,\"internalType\":\"address\",\"name\":\"to\",\"type\":\"address\"},{\"indexed\":false,\"internalType\":\"uint256[]\",\"name\":\"ids\",\"type\":\"uint256[]\"},{\"indexed\":false,\"internalType\":\"uint256[]\",\"name\":\"values\",\"type\":\"uint256[]\"}],\"name\":\"TransferBatch\",\"type\":\"event\"},{\"anonymous\":false,\"inputs\":[{\"indexed\":true,\"internalType\":\"address\",\"name\":\"operator\",\"type\":\"address\"},{\"indexed\":true,\"internalType\":\"address\",\"name\":\"from\",\"type\":\"address\"},{\"indexed\":true,\"internalType\":\"address\",\"name\":\"to\",\"type\":\"address\"},{\"indexed\":false,\"internalType\":\"uint256\",\"name\":\"id\",\"type\":\"uint256\"},{\"indexed\":false,\"internalType\":\"uint256\",\"name\":\"value\",\"type\":\"uint256\"}],\"name\":\"TransferSingle\",\"type\":\"event\"},{\"anonymous\":false,\"inputs\":[{\"indexed\":false,\"internalType\":\"string\",\"name\":\"value\",\"type\":\"string\"},{\"indexed\":true,\"internalType\":\"uint256\",\"name\":\"id\",\"type\":\"uint256\"}],\"name\":\"URI\",\"type\":\"event\"},{\"inputs\":[{\"internalType\":\"address\",\"name\":\"account\",\"type\":\"address\"},{\"internalType\":\"uint256\",\"name\":\"id\",\"type\":\"uint256\"}],\"name\":\"balanceOf\",\"outputs\":[{\"internalType\":\"uint256\",\"name\":\"\",\"type\":\"uint256\"}],\"stateMutability\":\"view\",\"type\":\"function\"},{\"inputs\":[{\"internalType\":\"address[]\",\"name\":\"accounts\",\"type\":\"address[]\"},{\"internalType\":\"uint256[]\",\"name\":\"ids\",\"type\":\"uint256[]\"}],\"name\":\"balanceOfBatch\",\"outputs\":[{\"internalType\":\"uint256[]\",\"name\":\"\",\"type\":\"uint256[]\"}],\"stateMutability\":\"view\",\"type\":\"function\"},{\"inputs\":[{\"internalType\":\"address\",\"name\":\"owner\",\"type\":\"address\"},{\"internalType\":\"uint256\",\"name\":\"tokenId\",\"type\":\"uint256\"},{\"internalType\":\"uint256\",\"name\":\"amount\",\"type\":\"uint256\"}],\"name\":\"burnToken\",\"outputs\":[],\"stateMutability\":\"nonpayable\",\"type\":\"function\"},{\"inputs\":[{\"internalType\":\"address\",\"name\":\"owner\",\"type\":\"address\"},{\"internalType\":\"uint256[]\",\"name\":\"tokenIds\",\"type\":\"uint256[]\"},{\"internalType\":\"uint256[]\",\"name\":\"amounts\",\"type\":\"uint256[]\"}],\"name\":\"burnTokenBatch\",\"outputs\":[],\"stateMutability\":\"nonpayable\",\"type\":\"function\"},{\"inputs\":[{\"internalType\":\"uint256\",\"name\":\"tokenId\",\"type\":\"uint256\"}],\"name\":\"getTokenInfo\",\"outputs\":[{\"internalType\":\"string\",\"name\":\"\",\"type\":\"string\"}],\"stateMutability\":\"view\",\"type\":\"function\"},{\"inputs\":[{\"internalType\":\"address\",\"name\":\"owner\",\"type\":\"address\"}],\"name\":\"getUserTokens\",\"outputs\":[{\"internalType\":\"uint256[]\",\"name\":\"\",\"type\":\"uint256[]\"}],\"stateMutability\":\"view\",\"type\":\"function\"},{\"inputs\":[{\"internalType\":\"address\",\"name\":\"account\",\"type\":\"address\"},{\"internalType\":\"address\",\"name\":\"operator\",\"type\":\"address\"}],\"name\":\"isApprovedForAll\",\"outputs\":[{\"internalType\":\"bool\",\"name\":\"\",\"type\":\"bool\"}],\"stateMutability\":\"view\",\"type\":\"function\"},{\"inputs\":[{\"internalType\":\"address\",\"name\":\"owner\",\"type\":\"address\"},{\"internalType\":\"uint256\",\"name\":\"tokenId\",\"type\":\"uint256\"},{\"internalType\":\"uint256\",\"name\":\"amount\",\"type\":\"uint256\"}],\"name\":\"mintToken\",\"outputs\":[],\"stateMutability\":\"nonpayable\",\"type\":\"function\"},{\"inputs\":[{\"internalType\":\"address\",\"name\":\"owner\",\"type\":\"address\"},{\"internalType\":\"uint256[]\",\"name\":\"tokenIds\",\"type\":\"uint256[]\"},{\"internalType\":\"uint256[]\",\"name\":\"amounts\",\"type\":\"uint256[]\"}],\"name\":\"mintTokenBatch\",\"outputs\":[],\"stateMutability\":\"nonpayable\",\"type\":\"function\"},{\"inputs\":[{\"internalType\":\"address\",\"name\":\"from\",\"type\":\"address\"},{\"internalType\":\"address\",\"name\":\"to\",\"type\":\"address\"},{\"internalType\":\"uint256[]\",\"name\":\"ids\",\"type\":\"uint256[]\"},{\"internalType\":\"uint256[]\",\"name\":\"amounts\",\"type\":\"uint256[]\"},{\"internalType\":\"bytes\",\"name\":\"data\",\"type\":\"bytes\"}],\"name\":\"safeBatchTransferFrom\",\"outputs\":[],\"stateMutability\":\"nonpayable\",\"type\":\"function\"},{\"inputs\":[{\"internalType\":\"address\",\"name\":\"from\",\"type\":\"address\"},{\"internalType\":\"address\",\"name\":\"to\",\"type\":\"address\"},{\"internalType\":\"uint256\",\"name\":\"id\",\"type\":\"uint256\"},{\"internalType\":\"uint256\",\"name\":\"amount\",\"type\":\"uint256\"},{\"internalType\":\"bytes\",\"name\":\"data\",\"type\":\"bytes\"}],\"name\":\"safeTransferFrom\",\"outputs\":[],\"stateMutability\":\"nonpayable\",\"type\":\"function\"},{\"inputs\":[{\"internalType\":\"address\",\"name\":\"operator\",\"type\":\"address\"},{\"internalType\":\"bool\",\"name\":\"approved\",\"type\":\"bool\"}],\"name\":\"setApprovalForAll\",\"outputs\":[],\"stateMutability\":\"nonpayable\",\"type\":\"function\"},{\"inputs\":[{\"internalType\":\"uint256\",\"name\":\"tokenId\",\"type\":\"uint256\"},{\"internalType\":\"string\",\"name\":\"tokenInfo\",\"type\":\"string\"}],\"name\":\"setTokenInfo\",\"outputs\":[],\"stateMutability\":\"nonpayable\",\"type\":\"function\"},{\"inputs\":[{\"internalType\":\"bytes4\",\"name\":\"interfaceId\",\"type\":\"bytes4\"}],\"name\":\"supportsInterface\",\"outputs\":[{\"internalType\":\"bool\",\"name\":\"\",\"type\":\"bool\"}],\"stateMutability\":\"view\",\"type\":\"function\"},{\"inputs\":[{\"internalType\":\"address\",\"name\":\"owner\",\"type\":\"address\"},{\"internalType\":\"address\",\"name\":\"to\",\"type\":\"address\"},{\"internalType\":\"uint256\",\"name\":\"tokenId\",\"type\":\"uint256\"},{\"internalType\":\"uint256\",\"name\":\"amount\",\"type\":\"uint256\"}],\"name\":\"transferToken\",\"outputs\":[],\"stateMutability\":\"nonpayable\",\"type\":\"function\"},{\"inputs\":[{\"internalType\":\"uint256\",\"name\":\"\",\"type\":\"uint256\"}],\"name\":\"uri\",\"outputs\":[{\"internalType\":\"string\",\"name\":\"\",\"type\":\"string\"}],\"stateMutability\":\"view\",\"type\":\"function\"}]";

    @Value("${chain.para.cAddr}")
    private String contractAddr;

    /**
     * 主动查询当前流水号对应的版权审批结果
     *
     * @param serial_code
     */
    @RabbitListener(queues = RabbitMQConfig.COPYRIGHT_QUEUE)
    public void listerCopyrightApply(String serial_code, Message message, Channel channel) throws IOException {
        log.info("监听者接收消息, 流水号为: {}", serial_code);
        long deliveryTag = message.getMessageProperties().getDeliveryTag();
        CopyrightQueryRequest request = new CopyrightQueryRequest(serial_code, copyrightProperties.getAppId());
        TreeMap<String, String> map = this.beanToMap(request, copyrightProperties.getAppSecret());
        String sign = CopyrightSignUtil.generateSign(map);
        request.setSign(sign);
        String json = JSONUtil.toJsonStr(request);
        log.info("查询审核结果, 当前流水号为:{} , 请求参数为 : {}", serial_code, json);
        String response = HttpUtil.post(copyrightProperties.getQueryUrl(), json, 10000);
        log.info("查询审核结果接口响应内容: {}", response);
        if (StringUtils.isBlank(response)) {
            // 接口无响应, 把消息丢进延时队列, 稍后重试
            rabbitTemplate.convertAndSend(RabbitMQConfig.COPYRIGHT_DIRECT, "copyright.apply", serial_code);
            channel.basicAck(deliveryTag, false);
            log.error("版权局查询存证接口无响应");
            return;
        }
        CopyrightQueryResponse copyrightResponse = JsonUtil.toObject(response, CopyrightQueryResponse.class);
        if (copyrightResponse == null) {
            // 接口无响应,手动签收, 把消息丢进延时队列, 稍后重试
            rabbitTemplate.convertAndSend(RabbitMQConfig.COPYRIGHT_DIRECT, "copyright.apply", serial_code);
            channel.basicAck(deliveryTag, false);
            log.error("版权局查询存证接口无响应");
            return;
        }
        if (!copyrightResponse.isSuccess()) {
            // 接口请求失败,手动签收, 把消息丢进延时队列, 稍后重试
            rabbitTemplate.convertAndSend(RabbitMQConfig.COPYRIGHT_DIRECT, "copyright.apply", serial_code);
            channel.basicAck(deliveryTag, false);
            log.error("接口响应失败, 失败原因:{}", copyrightResponse.getMessage());
            return;
        }
        try {
            CopyrightApply copyrightApply = copyrightApplyService.getBySerialNum(serial_code);
            //审核状态（0：未提交，1：待审核，2：同意，3：驳回，4：已提交，5：待终审，6：审核通过）
            if (!copyrightResponse.getAudit_status().equals(3) && !copyrightResponse.getAudit_status().equals(6)) {
                if (copyrightResponse.getAudit_status().equals(1)) {
                    // 待审核
                    copyrightApply.setRegisterState(CopyrightApplyState.EXAMINE.getCode());
                } else if (copyrightResponse.getAudit_status().equals(5)) {
                    // 待终审
                    copyrightApply.setRegisterState(CopyrightApplyState.FINAL_JUDGMENT.getCode());
                }
                // 审核还未完成, 需要把当前流水号丢进死信队列, 6小时后再去查询
                rabbitTemplate.convertAndSend(RabbitMQConfig.COPYRIGHT_DIRECT, "copyright.apply", serial_code);
            } else {

                if (copyrightResponse.getAudit_status().equals(6)) {
                    // 审核成功
                    String nftHash = copyrightApply.getNftHash();
                    copyrightApply.setEvidenceDate(DateUtil.parse(copyrightResponse.getApply_time(), "yyyy-MM-dd"));
                    copyrightApply.setRegisterCode(copyrightResponse.getCertificate().getRegister_code());
                    copyrightApply.setRegisterState(CopyrightApplyState.SUCCEEDED.getCode());
                    Nft nft = nftService.getByNftHash(nftHash);
                    String tokenInfo = paraChainClient.evmGetTokenInfo(abi, contractAddr, nft.getTokenId(), nft.getPublishAddress());
                    NftVo detail = nftService.getDetail(nftHash);
                    String newTokenInfo = appendTokenInfo(tokenInfo, detail.getCopyright());
                    log.info("追加tokenInfo, 新生成后的tokenInfo为: {}", newTokenInfo);
                    String hash = paraChainClient.evmSetTokenInfo(abi, contractAddr, 0, nft.getTokenId(), newTokenInfo, nft.getPublishAddress(), null, true);
                    if (StringUtils.isBlank(hash)) {
                        log.error("nft追加tokenInfo失败,当前流水号为: 【{}】,交易哈希:【{}】", serial_code, hash);
                        rabbitTemplate.convertAndSend(RabbitMQConfig.COPYRIGHT_DIRECT, "copyright.apply", serial_code);
                        return;
                    }
                    // 发送交易hash到死信队列,5秒后查询真实哈希,然后更新存证hash
                    EvidenceHashMessage msg = new EvidenceHashMessage(serial_code, hash);
                    rabbitTemplate.convertAndSend(RabbitMQConfig.COPYRIGHT_DIRECT, "evidence.hash.query", msg);
//                    String realTxHash = paraChainClient.getRealTxHashFromGrp(hash);
//                    log.info("真实哈希为: {}", realTxHash);
//                    copyrightApply.setEvidenceHash(realTxHash);
                } else {
                    // 审核失败
                    copyrightApply.setRejectReason(copyrightResponse.getRemark());
                    copyrightApply.setRegisterState(CopyrightApplyState.FAILED.getCode());
                }
                copyrightApplyService.updateById(copyrightApply);
            }
        } catch (Exception e) {
            // 如果上链失败或者更新状态失败, 那么就把消息重新丢入延时队列,稍后重试
            rabbitTemplate.convertAndSend(RabbitMQConfig.COPYRIGHT_DIRECT, "copyright.apply", serial_code);
            log.error(e.getMessage(), e);
        } finally {
            channel.basicAck(deliveryTag, false);
        }

    }


    /**
     * 更新存证hash
     */
    @RabbitListener(queues = RabbitMQConfig.EVIDENCE_HASH_QUEUE)
    public void listenEvidenceHash(EvidenceHashMessage evidenceHashMessage, Message message, Channel channel) throws IOException {
        long deliveryTag = message.getMessageProperties().getDeliveryTag();
        try {
            String hash = evidenceHashMessage.getHash();
            String serial_code = evidenceHashMessage.getSerial_code();
            String realTxHash = paraChainClient.getRealTxHashFromGrp(hash);
            if (StringUtils.isBlank(realTxHash)) {
                log.error("存证hash查询为空, 交易hash为: {}, 流水号为: {}", hash, serial_code);
                // 查询失败,拒绝签收
                channel.basicReject(deliveryTag, true);
                return;
            }
            CopyrightApply apply = copyrightApplyService.getBySerialNum(serial_code);
            apply.setEvidenceHash(realTxHash);
            copyrightApplyService.updateById(apply);
            // 手动签收
            channel.basicAck(deliveryTag, false);
        } catch (Exception e) {
            log.error(e.getMessage(), e);
            // 发生异常, 拒收消息
            channel.basicReject(deliveryTag, true);
        }


    }

    private TreeMap<String, String> beanToMap(CopyrightQueryRequest request, String app_secret) {
        TreeMap<String, String> map = JSONUtil.toBean(JSONUtil.toJsonStr(request), new TypeReference<TreeMap<String, String>>() {
        }, true);
        map.put("app_secret", app_secret);
        return map;
    }

    private String appendTokenInfo(String tokenInfo, NftVo.Copyright copyright) {
        String[] split = tokenInfo.split("&");
        TreeMap<String, String> map = new TreeMap<>();
        for (String s : split) {
            String[] arr = s.split("=");
            String key = arr[0];
            String value;
            if (arr.length < 2) {
                value = "";
            } else {
                value = arr[1];
            }
            map.put(key, value);
        }
        map.put("evidencer", copyright.getEvidencer());
        map.put("opusName", copyright.getOpusName());
        map.put("opusType", copyright.getOpusType());
        map.put("author", copyright.getAuthor());
        map.put("copyrightOwner", copyright.getCopyrightOwner());
        map.put("registerCode", copyright.getRegisterCode());
        StringBuilder stringBuilder = new StringBuilder();
        for (Map.Entry<String, String> entry : map.entrySet()) {
            String value = entry.getValue();
            String key = entry.getKey();
            stringBuilder.append(key).append("=").append(value).append("&");
        }
        return stringBuilder.deleteCharAt(stringBuilder.length() - 1).toString();
    }


}
